#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <readline/readline.h>
#include <getopt.h>
#include "cline.h"
#include "strcase.h"

FILE * os = stdout;
FILE * is = stdin;
char * error_noversion = "No version loaded";
char * error_nobook = "No book loaded";

static void message(char * mesg)
{
	if (mesg != NULL)
		fprintf(stderr, "%s\n", mesg);
}

static void showVerse(Book * book, AbsVerse ver)
	{
	register char * ptr;
	fwrite(book->data + book->verses[ver], verseLength(book->data + book->verses[ver]), 1, os);
	}

Bool checkVersion(ClineContext * con)
{
	if (con->context.cVersion == NULL)
		{
		message(error_noversion);
		return NO;
		}
	return YES;
}

Bool checkBook(ClineContext * con)
{
	if (con->context.cVersion->cBook == NULL)
		{
		message(error_nobook);
		return NO;
		}
	return YES;
}

int clineRunArgs(ClineContext * con)
	{
	if (!checkVersion(con))
		return 1;
	while (c_showNext(con, con->range, con->searchFor))
		;
	return 0;
	}

c_showNext(ClineContext * con, Range * range, char * search)
	{
	ErrorBuf errbuf;
	Bool nBook;
	
	if ((con->verPosition = searchNextVersion(con->context.cVersion, con->verPosition, range, search, errbuf, NO, &nBook)) != MAT_NOMATCH)
		{
		if (0 <= con->verPosition)
			{
			showVerse(con->context.cVersion->cBook, con->verPosition);
			con->verPosition++;
			return YES;
			}
		else
			message(errbuf);
		}
	return NO;
	}

char * lastline = NULL;

#ifdef GNU_READLINE

ClineContext * cgcon = NULL;


/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
char * command_generator (char * text, int state)
	{
	static int list_index, len;
	static Bool bookMarks;
	char *name;
	
	/* If this is a new word to complete, initialize now.  This includes
	   saving the length of TEXT for efficiency, and initializing the index
	   variable to 0. */
	if (!state)
		{
		list_index = 0;
		len = strlen (text);
		bookMarks = NO;
		}
	
	/* Return the next name which partially matches from the command list. */
	if (!bookMarks)
		{
		while (list_index < cgcon->context.cVersion->numBooks)
			{
			name = cgcon->context.cVersion->books[list_index].title; 
			list_index++;
			
			if (strncmp (name, text, len) == 0)
				return strdup(name);
			}
		bookMarks = YES;
		list_index = 0;
		}
	
	while (list_index < bmarklist->num)
		{
		name = bmarklist->bookmarks[list_index].title;
		list_index++;
		
		if (strncmp (name, text, len) == 0)
			return strdup(name);
		}
	
	/* If no names matched, then return NULL. */
	return ((char *)NULL);
	}

/* Attempt to complete on the contents of TEXT.  START and END show the
   region of TEXT that contains the word to complete.  We can use the
   entire line in case we want to do some simple parsing.  Return the
   array of matches, or NULL if there aren't any. */
char ** fileman_completion (char * text, int start, int end)
	{
	char **matches;
	
	matches = (char **)NULL;
	
	/* If this word is at the start of the line, then it is a command
	   to complete.  Otherwise it is the name of a file in the current
	   directory. */
	matches = completion_matches (text, command_generator);
	
	return (matches);
	}

initialise_readline()
	{
	static char * extra_delim = "-,", * old_delim;
	/* Allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "Bible";
	
	/* Tell the completer that we want a crack first. */
	rl_attempted_completion_function = (CPPFunction *)fileman_completion;
	old_delim = rl_basic_word_break_characters;
	rl_basic_word_break_characters = (char *)malloc(strlen(rl_basic_word_break_characters) + strlen(extra_delim) + 1);
	strcpy(rl_basic_word_break_characters, old_delim);
	strcat(rl_basic_word_break_characters, extra_delim);
	}


char * getline(ClineContext * con, char * prompt, Bool emptyDoLast)
	{
	char * rtn;
	
	cgcon = con;
	rtn = readline(prompt);
	if (rtn != NULL && rtn[0] == EOS && lastline != NULL && emptyDoLast)
		rtn = strdup(lastline);
	if (rtn != NULL && rtn[0] != EOS)
		{
		if (lastline == NULL || strcmp(lastline, rtn) != EQUAL)
			add_history(rtn);
		}
	return rtn;
	}

void freeline(char * line)
	{
	if (lastline != NULL)
		free(lastline);
	lastline = line;
	}

#else /* GNU_READLNE */
char * getline(ClineContext * con, char * prompt, Bool emptyDoLast)
	{
	char buf[1024];

	printf("%s", prompt);
	fflush(stdout);
	if (fgets(buf, sizeof(buf), is) == NULL)
		return NULL;
	return strdup(buf);
	}

void freeline(char * line)
	{
	free(line);
	}
#endif /* GNU_READLNE */

char * getstdline(ClineContext * con)
	{
	char prompt[256+1];
	char * spec = "%s!%r[%v.%b] ";
	char * vername = "";
	char * bookname = "";
	char * range = "";
	char * search = "";
	int count, pcount = 0, len;

	if (con->context.cVersion != NULL )
		vername = con->context.cVersion->title;
	if (con->context.cVersion != NULL && con->context.cVersion->cBook != NULL)
		bookname = con->context.cVersion->cBook->title;
	if (con->rangeString != NULL)
		range = con->rangeString;
	if (con->searchFor != NULL)
		search = con->searchFor;
	for (count = 0; spec[count] != EOS; count++)
	{
		if (spec[count] == '%')
		{
			switch (spec[count+1])
			{
			case 'v':
				len = strlen(vername);
				memcpy(&prompt[pcount], vername, len);
				pcount += len;
				count++;
				break;
			case 'b':
				len = strlen(bookname);
				memcpy(&prompt[pcount], bookname, len);
				pcount += len;
				count++;
				break;
			case 'r':
				len = strlen(range);
				memcpy(&prompt[pcount], range, len);
				pcount += len;
				count++;
				break;
			case 's':
				len = strlen(search);
				memcpy(&prompt[pcount], search, len);
				pcount += len;
				count++;
				break;
			case EOS:
				prompt[pcount++] = '%';
				break;
			default:
				prompt[pcount++] = '%';
				count++;
				break;
			}
		}
		else
			prompt[pcount++] = spec[count];
	}
	prompt[pcount] = EOS;
	/*
	sprintf(prompt, "%s[%s.%s] ", con->rangeString, vername, bookname);
	*/
	return getline(con, prompt, YES);
	}


char * getTok(register char * line, register char * tok)
	{
	char * beg = tok;

	skipwhite(line);
	while (*line != EOS && !isspace(*line))
		*tok++ = *line++;
	*tok = EOS;
	if (*beg == EOS)
		return NULL;
	else
		return line;
	}

#define MAX_COMMAND_LEN 1024
#define MAX_COMMAND_ARGS 256

char ** getArgvSVec(char * line)
	{
	char command[MAX_COMMAND_LEN];
	char ** rtn, ** ptr;

	ptr = rtn = (char **)malloc(MAX_COMMAND_ARGS * sizeof(char *));
	while ((line = getTok(line, command)) != NULL)
		*ptr++ = strdup(command);
	*ptr = NULL;
	return rtn;
	}

char * promptField(ClineContext * con, char * prompt)
	{
	char * rtn;
	
#ifdef NEVER
	while ((rtn = getline(con, prompt, NO)) != NULL)
		{
		if (rtn[0] == EOS)
			free(rtn);
		else
			break;
		}
#endif
	rtn = getline(con, prompt, NO);
	return rtn;
	}

void c_loadV(char ** command, ClineContext * con)
	{
	if (command[1] == NULL)
		if ((command[1] = promptField(con, "Version name: ")) == NULL)
			return;
	command[2] = NULL;

	
	if ((con->context.cVersion = findVersion(command[1])) == NULL)
		{
		char ** vec, ** ptr;
		int num;
		
		fprintf(stderr, "Bad Version '%s'\n", command[1]);
		fprintf(stderr, "Valid versions are: ");
		if ((vec = ptr = getVersionListSVec(&num)) == NULL)
			fprintf(stderr, "\nOoops! You havn't installed any versions!\n");
		else
			{
			while (*ptr != NULL)
				fprintf(stderr, "%s ", *ptr++);
			fprintf(stderr, "\n");
			freevector(vec);
			}
		}
	}


void c_loadB(char ** command, ClineContext * con)
	{
	if (!checkVersion(con))
		return;
	if (command[1] == NULL)
		if ((command[1] = promptField(con, "Book name: ")) == NULL)
			return;
	command[2] = NULL;
	
	if (loadBookAsCurrent(con->context.cVersion, command[1]))
		{
		con->verPosition = 0;
		showVerse(con->context.cVersion->cBook, con->verPosition);
		}
	else
		{
		char ** vec, ** ptr;
		int num;
		
		fprintf(stderr, "Bad book name '%s'\n", command[1]);
		fprintf(stderr, "Valid books are: ");
		vec = ptr = getBookContentsSVec(con->context.cVersion->title, &num);
		while (*ptr != NULL)
			fprintf(stderr, "%s ", *ptr++);
		fprintf(stderr, "\n");
		freevector(vec);
		}
	}

char * END = "-- END --\n";

void c_next(char ** command, ClineContext * con)
	{
	if (!checkVersion(con))
		return;
	else if (!checkBook(con))
		return;
	else
		{
		if (con->verPosition == con->context.cVersion->cBook->numVerses - 1)
			{
			int book;
			
			if (con->context.cVersion->numBooks - 1 <= (book = findBookIndex(con->context.cVersion, con->context.cVersion->cBook->title)))
				{
				puts(END);
				return;
				}
			else
				{
				loadBookAsCurrent(con->context.cVersion, con->context.cVersion->books[book+1].title);
				con->verPosition = 0;
				}
			}
		else
			con->verPosition++;
		showVerse(con->context.cVersion->cBook, con->verPosition);
		}
	}

void c_prev(char ** command, ClineContext * con)
	{
	if (!checkVersion(con))
		return;
	else if (!checkBook(con))
		return;
	else
		{
		if (con->verPosition == 0)
			{
			int book;
			
			if ((book = findBookIndex(con->context.cVersion, con->context.cVersion->cBook->title)) <= 0)
				{
				printf(END);
				return;
				}
			else
				{
				loadBookAsCurrent(con->context.cVersion, con->context.cVersion->books[book-1].title);
				con->verPosition = con->context.cVersion->cBook->numVerses - 1;
				}
			}
		else
			con->verPosition--;
		showVerse(con->context.cVersion->cBook, con->verPosition);
		}
	}

void c_goto(char ** command, ClineContext * con)
	{
	AbsVerse res;
	Bool nBook;
	if (!checkVersion(con))
		return;
	else if (!checkBook(con))
		return;
	if (command[1] == NULL)
		if ((command[1] = promptField(con, "Goto verse: ")) == NULL)
			return;
	command[2] = NULL;
	if ((res = gotoVerseString(con->context.cVersion, command[1], &nBook)) == noVerseNum)
		fprintf(stderr, "Bad verse description\n");
	else
		{
		con->verPosition = res;
		showVerse(con->context.cVersion->cBook, con->verPosition);
		}
	}

void c_range(char ** command, ClineContext * con)
	{
	int eaten;

	if (command[1] == NULL)
		if ((command[1] = promptField(con, "Default Range: ")) == NULL)
			return;
	if (con->rangeString != NULL)
		free(con->rangeString);
	con->rangeString = strdup(command[1]);
	con->range = parseVerse(con->rangeString, &eaten);
	}

void c_search(char ** command, ClineContext * con)
	{

	if (command[1] == NULL)
		if ((command[1] = promptField(con, "Search for: ")) == NULL)
			return;
	if (con->searchFor != NULL)
		free(con->searchFor);
	con->searchFor = strdup(command[1]);
	con->verPosition = 0; /* is this right? */
	}


void c_show(char ** command, ClineContext * con)
	{
	if (!checkVersion(con))
		return;
	if (command[1] == NULL || strcmp(command[1], "all") == 0)
		while (c_showNext(con, con->range, con->searchFor))
			;
	else
		{
			if (strcmp(command[1], "next") == 0)
				c_showNext(con, con->range, con->searchFor);
			else if (strcmp(command[1], "prev") == 0)
				message("not implemented yet");
			else
				message("usage: show [all|next|prev]");
		}
	}

c_help()
{
	printf("%s", 
	"? - help\n"
	"loadv - load a version\n"
	"loadb - load a book\n"
	"> - forward\n"
	"< - back\n"
	"goto - goto verse\n"
	"range - set default range\n"
	"search - search for text\n"
	"show [next] - show search results\n"
	"q - quit\n"
	);
}

Bool parseExecuteLine(ClineContext * con, char * line)
	{
	char ** command;

	command = getArgvSVec(line);

	STRSWITCH(command[0]);
	STRCASE(NULL);
	STRCASE("?");
	c_help();
	STRCASE("loadv");
	c_loadV(command, con);
	STRCASE("loadb");
	c_loadB(command, con);
	STRCASE(">");
	c_next(command, con);
	STRCASE("<");
	c_prev(command, con);
	STRCASE("goto");
	c_goto(command, con);
	STRCASE("range");
	c_range(command, con);
	STRCASE("search");
	c_search(command, con);
	STRCASE("show");
	c_show(command, con);
	STRCASE("q");
	return NO;
	STRDEFAULT;
	fprintf(stderr, "Unknown command '%s'. Type ? for help\n", command[0]);
	STRENDSWITCH;
	freevector(command);
	return YES;
	}

int clineRunInteractive(ClineContext * con)
	{
	char * line;
	Bool cont = YES;

	while ((line = getstdline(con)) != NULL)
		{
		if (line[0] == EOS)
			{
			if (lastline != NULL)
				cont = parseExecuteLine(con, lastline);
			free(line);
			}
		else
			{
			cont = parseExecuteLine(con, line);
			freeline(line);
			}
			if (!cont)
				break;
		}
	return 0;
	}

int clinerun(ClineContext * con)
	{
	if (con->interactive)
		return clineRunInteractive(con);
	else
		return clineRunArgs(con);
	}


void usage()
{
	fprintf(stderr, "usage: %s [-s searchtext] [-r range] [-v version] [-i(interactive)]\n", 
			progname);
	exit(1);
}

ClineContext * loadcline(int argc, char **argv)
{
	int optchar;
	int errcount = 0;
	int eaten;
	char * version = NULL;
	ClineContext * con;

#ifdef GNU_READLINE
	initialise_readline();
#endif;		
	con = (ClineContext *)malloc(sizeof (ClineContext));
	con->context.run = (int (*)(Context *))clinerun;
	con->context.cVersion = NULL;
	con->searchFor = NULL;
	con->rangeString = NULL;
	con->interactive = NO;
	con->range = NULL;
	con->verPosition = 0;
	progname = argv[0];

	while ((optchar = getopt(argc, argv, "?s:r:v:i")) != -1)
		switch (optchar)
			{
		case 's': con->searchFor = strdup(optarg); break;
		case 'r': con->rangeString = strdup(optarg); break;
		case 'v': version = strdup(optarg); break;
		case 'i': con->interactive = YES; break;
		case '?': errcount++; break;
		default: errcount++ ; break;
			}
	argv += optind;
	argc -= optind;
	if (errcount)
		usage();
	
	if (version != NULL)
		con->context.cVersion = loadVersion(version);
	if (con->searchFor == NULL)
		con->interactive = YES;
	if (con->rangeString != NULL)
		{
		con->range = parseVerse(con->rangeString, &eaten);
		if (eaten != strlen(con->rangeString))
			fprintf(stderr, "Trailing garbage on range string ignored\n");
		}
	return con;
}
